/***************************************************************
 * stack.hh ---- header file for proj1.cc and stack.cc         *
 *                                                             *
 ***************************************************************/

#ifndef STACK_HH

#define STACK_HH

#include <stdio.h>   // NULL
// link list structure
struct link_list {   
  char value;
  link_list *next;
};

// Stack of integers
class Stack {	
  public:
	Stack() { top = NULL; }	   // constructor
	void push(char element);   // push operation
	char pop(void); 	   // pop  operation
	void show(); 	           // show operation 
	link_list * scan(char element); // scan operation
	void copy(Stack& to); 	   // copy operation
	~Stack() { while (pop()); }     // distructor
  private:
	link_list *top;
};

#endif STACK_HH

/***************************************************************
 * stack.cc ---- modul of information cluster for stack class  *
 *              (C++ version)                                  *
 ***************************************************************/

//#include "stack.hh"  // stack header file
#include <fstream.h> // ifstream

// ----<>----//
/* push operation */
void Stack :: push(char element)
{
  link_list *tmp = new link_list;

  tmp->value = element;
  tmp->next = top;
  top = tmp;

  return;
}

// ----<char Stack :: pop()>----//
/* pop operation */
char Stack :: pop()
{
  if (top == NULL) return NULL;

  link_list *tmp = top;
  char ch = top->value; 
  top = top->next;
  delete tmp;

  return ch;
}

// ----<void Stack :: show()>----//
/* display on stdout all elements in a stack (in the order from top to bottom) */
void Stack :: show() 
{
  link_list *tmp = top;  
  while (tmp) {
    cout << tmp->value;
    tmp = tmp->next;
  };

  return;
}

// ----<link_list * Stack :: scan(char element)>----//
link_list * Stack :: scan(char element)
{
  link_list *tmp = top;  
  while (tmp && (tmp->value != element)) tmp = tmp->next;
  
  return (tmp);
} 

// ----<void Stack :: copy(Stack& to)>----//
/* copy all elements in a stack to another stack */
void Stack :: copy(Stack& to)
{
  if (this->top == to.top) return; // return if copy to itself  
  while (to.pop());        // free any space left 

  if (this->top == NULL) 
    return;

  link_list *tmp_from = this->top;
  link_list *tmp_to = to.top = new link_list;
  tmp_to->value = tmp_from->value;
  link_list *tmp = tmp_to;
  tmp_from = tmp_from->next;

  // copy elementwise from top to bottom
  while (tmp_from != NULL) {
    tmp->next = tmp_to = new link_list;
    (tmp = tmp_to)->value = tmp_from->value;
    tmp_from = tmp_from->next;
  };
  tmp->next = NULL;

  return;
}
    
 
/***************************************************************
 * proj1.cc ---- main program to test the information cluster  *
 *               for stack class (C++ version)                 *
 ***************************************************************/

//#include <fstream.h> // ifstream
//#include "stack.hh"  // class Stack

extern "C" exit(int); // exit function

#define default_file "proj1.txt" // default file name used for input 

/* function prototypes */
void read_into_stack(ifstream fp, Stack &stack_id);
void write_onto_stack(Stack &stack_id1,Stack &stack_id2);
void error(char *s1, char *s2 = "");

// ---- <void error(char *s1, char *s2 = "")> ---- //
/* error handler */
void error(char *s1, char *s2 = "")
{
  cerr << s1 << ' ' << s2 << '\n';
  exit(1);
}

// <main program> //
/* test the information cluster for stack class
   tested member functions of class Stack:
   . stack constructor/destructor . stack copier
   . show all elements in a stack (from top to bottom) */
main(int argc, char *argv[])
{
  // open file specified on command line or the defaulted one
  ifstream fp((argc>1) ? argv[1] : default_file);
 
  // check i/o status
  if (!fp) {
    if (argc > 1) {
       cerr << "cannot open file: " << argv[1] << "\n";
       cout << "now try to open the default file: " << default_file << "\n";
       // open the default one
       ifstream gp;      // looks silly, but there is a problem on cat system
       gp.open(default_file); // this is a tempory solution (gp is not necessary)
       if (!gp) error("cannot open file: ",default_file);
       argc = 1;
       fp = gp;
     }
    else error("cannot open file: ", default_file);
  };
  
  cout << "This is file " << ((argc == 1) ? default_file : argv[1])
    << "\nProcessed by " <<  argv[0] << "\n";

  // Create stack1,stack3 (definition and construction go togather)
  Stack stack1,stack3;  

  // read the stream in file into stack1
  read_into_stack(fp,stack1);

  // copy stack1 to stack 3
  stack1.copy(stack3);

  // print a centered line 
  cout << "\n--------------------- Reversed Stream -----------------------\n";

  // show reversed stream in stack1
  stack1.show();

  // Create stack2,stack4 (definition and construction go togather)
  Stack stack2, stack4;  

  // write the reversed stream from stack1 onto stack2 
  write_onto_stack(stack1,stack2); 

  // print a centered line
  cout << "\n--------------------- Normal Stream -------------------------\n";

  // show  stream in stack2
  stack2.show();

  cout << '\n'; // end newline 

  // copy stack2 to stack4
  stack2.copy(stack4);  

 // show once more using copied stacks
  cout << "\n------------------ Again Reversed Stream --------------------\n";
  stack3.show();
  cout << "\n------------------ Again Normal Stream ----------------------\n";
  stack4.show();
  cout << '\n'; 

  // stacks will be destroyed automatically when progam exits
  exit(0); 
}

// ---- <void read_into_stack(ifstream fp, Stack &stack_id)> ---- //
// read stream from a file into stack //
inline void read_into_stack(ifstream fp, Stack &stack_id)
{
  char ch;
  while (fp.get(ch)) stack_id.push(ch);

  return;
}

// ---- <void write_onto_stack(Stack &stack_id1,Stack& stack_id2)> ---- //
// move elements from a stack into another stack //
inline void write_onto_stack(Stack &stack_id1,Stack& stack_id2)
{
  char ch;
  while ((ch = stack_id1.pop()) != NULL) stack_id2.push(ch); 

  return;
}

